from __future__ import annotations


from abc import abstractmethod
from typing import (
    Iterator,
    Self,
    TypedDict,
    Unpack
)

import numpy as np

from ..constants.constants import (
    ORIGIN,
    PI
)
from ..constants.custom_typing import (
    ColorType,
    NP_3f8,
    NP_44f8,
    NP_x3f8
)
from ..lazy.lazy import Lazy
from ..lazy.lazy_object import LazyObject
from ..rendering.buffers.uniform_block_buffer import UniformBlockBuffer
from .animatable.actions import (
    Action,
    ConverterDescriptorParameters,
    DescriptiveAction
)
from .animatable.animatable import (
    Animatable,
    AnimatableActions,
    DynamicAnimatable
)
from .animatable.animation import (
    AnimateKwargs,
    Animation
)
from .arrays.model_matrix import ModelMatrix


class SetKwargs(TypedDict, total=False):
    # polymorphism variables
    color: ColorType
    opacity: float
    weight: float

    # MeshMobject
    ambient_strength: float
    specular_strength: float
    shininess: float

    # GraphMobject
    width: float


class Box(LazyObject):
    __slots__ = ()

    def __init__(
        self: Self,
        maximum: NP_3f8,
        minimum: NP_3f8
    ) -> None:
        super().__init__()
        self._maximum_ = maximum
        self._minimum_ = minimum

    @Lazy.variable()
    @staticmethod
    def _maximum_() -> NP_3f8:
        return NotImplemented

    @Lazy.variable()
    @staticmethod
    def _minimum_() -> NP_3f8:
        return NotImplemented

    @Lazy.property()
    @staticmethod
    def _centroid_(
        maximum: NP_3f8,
        minimum: NP_3f8
    ) -> NP_3f8:
        return (maximum + minimum) / 2.0

    @Lazy.property()
    @staticmethod
    def _radii_(
        maximum: NP_3f8,
        minimum: NP_3f8
    ) -> NP_3f8:
        return np.maximum((maximum - minimum) / 2.0, 1e-8)  # Avoid zero-divisions.

    def get(
        self: Self,
        direction: NP_3f8 = ORIGIN,
        buff: float | NP_3f8 = 0.0
    ) -> NP_3f8:
        return self._centroid_ + self._radii_ * direction + buff * direction

    def get_radii(
        self: Self
    ) -> NP_3f8:
        return self._radii_


class ModelActions(AnimatableActions):
    __slots__ = ()

    @DescriptiveAction.register(ConverterDescriptorParameters)
    @classmethod
    def set(
        cls: type[Self],
        dst: Model,
        broadcast: bool = True,
        **kwargs: Unpack[SetKwargs]
    ) -> Iterator[Animation]:
        for descriptor, descriptor_parameters in cls.set.iter_descriptor_items():
            if (input_value := kwargs.get(descriptor._name.strip("_"))) is None:
                continue
            converter = descriptor_parameters.converter
            target = converter(input_value)
            for sibling in dst._iter_siblings(broadcast=broadcast):
                if descriptor not in sibling._lazy_descriptors:
                    continue
                initial = descriptor.__get__(sibling)
                assert isinstance(initial, Animatable)
                yield from type(initial).interpolate.iter_animations(
                    dst=initial,
                    src_0=initial.copy(),
                    src_1=target
                )

    @Action.register()
    @classmethod
    def shift(
        cls: type[Self],
        dst: Model,
        vector: NP_3f8,
        mask: float | NP_3f8 = 1.0
    ) -> Iterator[Animation]:
        yield ModelShiftAnimation(
            model=dst,
            vector=vector,
            mask=mask * np.ones((3,))
        )

    @Action.register()
    @classmethod
    def move_to(
        cls: type[Self],
        dst: Model,
        target: Model,
        direction: NP_3f8 = ORIGIN,
        buff: float | NP_3f8 = 0.0,
        mask: float | NP_3f8 = 1.0
    ) -> Iterator[Animation]:
        yield from cls.shift.iter_animations(
            dst=dst,
            vector=target.box.get(direction) - dst.box.get(direction, buff),
            mask=mask
        )

    @Action.register()
    @classmethod
    def next_to(
        cls: type[Self],
        dst: Model,
        target: Model,
        direction: NP_3f8 = ORIGIN,
        buff: float | NP_3f8 = 0.0,
        mask: float | NP_3f8 = 1.0
    ) -> Iterator[Animation]:
        yield from cls.shift.iter_animations(
            dst=dst,
            vector=target.box.get(direction) - dst.box.get(-direction, buff),
            mask=mask
        )

    @Action.register()
    @classmethod
    def scale(
        cls: type[Self],
        dst: Model,
        factor: float | NP_3f8,
        about: Model | None = None,
        direction: NP_3f8 = ORIGIN,
        mask: float | NP_3f8 = 1.0
    ) -> Iterator[Animation]:
        if about is None:
            about = dst
        yield ModelScaleAnimation(
            model=dst,
            factor=factor * np.ones((3,)),
            about=about,
            direction=direction,
            mask=mask * np.ones((3,))
        )

    @Action.register()
    @classmethod
    def scale_about_origin(
        cls: type[Self],
        dst: Model,
        factor: float | NP_3f8,
        mask: float | NP_3f8 = 1.0
    ) -> Iterator[Animation]:
        yield from cls.scale.iter_animations(
            dst=dst,
            factor=factor,
            about=Model(),
            mask=mask
        )

    @Action.register()
    @classmethod
    def scale_to(
        cls: type[Self],
        dst: Model,
        target: Model,
        about: Model | None = None,
        direction: NP_3f8 = ORIGIN,
        mask: float | NP_3f8 = 1.0
    ) -> Iterator[Animation]:
        yield from cls.scale.iter_animations(
            dst=dst,
            factor=target.box.get_radii() / dst.box.get_radii(),
            about=about,
            direction=direction,
            mask=mask
        )

    @Action.register()
    @classmethod
    def rotate(
        cls: type[Self],
        dst: Model,
        rotvec: NP_3f8,
        about: Model | None = None,
        direction: NP_3f8 = ORIGIN,
        mask: float | NP_3f8 = 1.0
    ) -> Iterator[Animation]:
        if about is None:
            about = dst
        yield ModelRotateAnimation(
            model=dst,
            rotvec=rotvec,
            about=about,
            direction=direction,
            mask=mask * np.ones((3,))
        )

    @Action.register()
    @classmethod
    def rotate_about_origin(
        cls: type[Self],
        dst: Model,
        rotvec: NP_3f8,
        mask: float | NP_3f8 = 1.0
    ) -> Iterator[Animation]:
        yield from cls.rotate.iter_animations(
            dst=dst,
            rotvec=rotvec,
            about=Model(),
            mask=mask
        )

    @Action.register()
    @classmethod
    def flip(
        cls: type[Self],
        dst: Model,
        axis: NP_3f8,
        about: Model | None = None,
        direction: NP_3f8 = ORIGIN,
        mask: float | NP_3f8 = 1.0
    ) -> Iterator[Animation]:
        yield from cls.rotate.iter_animations(
            dst=dst,
            rotvec=axis / np.linalg.norm(axis) * PI,
            about=about,
            direction=direction,
            mask=mask
        )

    @Action.register()
    @classmethod
    def apply(
        cls: type[Self],
        dst: Model,
        matrix: NP_44f8,
        about: Model | None = None,
        direction: NP_3f8 = ORIGIN
    ) -> Iterator[Animation]:
        if about is None:
            about = dst
        yield ModelApplyAnimation(
            model=dst,
            matrix=matrix,
            about=about,
            direction=direction
        )


class Model(ModelActions, Animatable):
    __slots__ = ()

    @AnimatableActions.interpolate.register_descriptor()
    @Lazy.volatile()
    @staticmethod
    def _model_matrix_() -> ModelMatrix:
        return ModelMatrix()

    @AnimatableActions.interpolate.register_descriptor()
    @AnimatableActions.piecewise.register_descriptor()
    @Lazy.volatile(plural=True)
    @staticmethod
    def _proper_siblings_() -> tuple[Model, ...]:
        return ()

    @Lazy.property()
    @staticmethod
    def _model_uniform_block_buffer_(
        model_matrix__array: NP_44f8
    ) -> UniformBlockBuffer:
        return UniformBlockBuffer(
            name="ub_model",
            field_declarations=(
                "mat4 u_model_matrix",
            ),
            data_dict={
                "u_model_matrix": model_matrix__array.T
            }
        )

    @Lazy.property()
    @staticmethod
    def _local_sample_positions_() -> NP_x3f8:
        return np.zeros((0, 3))

    @Lazy.property()
    @staticmethod
    def _world_sample_positions_(
        model_matrix__array: NP_44f8,
        local_sample_positions: NP_x3f8,
    ) -> NP_x3f8:
        return ModelMatrix._apply_multiple(model_matrix__array, local_sample_positions)

    @Lazy.property()
    @staticmethod
    def _box_(
        world_sample_positions: NP_x3f8,
        proper_siblings__world_sample_positions: tuple[NP_x3f8, ...]
    ) -> Box:
        sample_positions = np.concatenate((
            world_sample_positions,
            *proper_siblings__world_sample_positions
        ))
        if not len(sample_positions):
            return Box(
                maximum=np.zeros((3,)),
                minimum=np.zeros((3,))
            )
        return Box(
            maximum=sample_positions.max(axis=0),
            minimum=sample_positions.min(axis=0)
        )

    def _iter_siblings(
        self: Self,
        *,
        broadcast: bool = True
    ) -> Iterator[Model]:
        yield self
        if broadcast:
            yield from self._proper_siblings_

    @property
    def box(
        self: Self
    ) -> Box:
        return self._box_

    def animate(
        self: Self,
        **kwargs: Unpack[AnimateKwargs]
    ) -> DynamicModel[Self]:
        return DynamicModel(self, **kwargs)


class DynamicModel[ModelT: Model](ModelActions, DynamicAnimatable[ModelT]):
    __slots__ = ()


class ModelAnimation(Animation):
    __slots__ = (
        "_pre_shift_matrix",
        "_post_shift_matrix",
        "_model_matrices"
    )

    def __init__(
        self: Self,
        model: Model,
        about: Model,
        direction: NP_3f8
    ) -> None:
        super().__init__()
        about_point = about.box.get(direction)
        self._pre_shift_matrix: NP_44f8 = ModelMatrix._matrix_from_shift(-about_point)
        self._post_shift_matrix: NP_44f8 = ModelMatrix._matrix_from_shift(about_point)
        self._model_matrices: dict[ModelMatrix, NP_44f8] = {
            sibling._model_matrix_: sibling._model_matrix_._array_
            for sibling in (model, *model._proper_siblings_)
        }

    @abstractmethod
    def _get_matrix(
        self: Self,
        alpha: float
    ) -> NP_44f8:
        pass

    def update(
        self: Self,
        alpha: float
    ) -> None:
        super().update(alpha)
        matrix = self._post_shift_matrix @ self._get_matrix(alpha) @ self._pre_shift_matrix
        for model_matrix, initial_model_matrix_array in self._model_matrices.items():
            model_matrix._array_ = matrix @ initial_model_matrix_array


class ModelShiftAnimation(ModelAnimation):
    __slots__ = (
        "_vector",
        "_mask"
    )

    def __init__(
        self: Self,
        model: Model,
        vector: NP_3f8,
        mask: NP_3f8
    ) -> None:
        super().__init__(
            model=model,
            about=Model(),
            direction=ORIGIN
        )
        self._vector: NP_3f8 = vector
        self._mask: NP_3f8 = mask

    def _get_matrix(
        self: Self,
        alpha: float
    ) -> NP_44f8:
        return ModelMatrix._matrix_from_shift(self._vector * (self._mask * alpha))


class ModelScaleAnimation(ModelAnimation):
    __slots__ = (
        "_factor",
        "_mask"
    )

    def __init__(
        self: Self,
        model: Model,
        factor: NP_3f8,
        about: Model,
        direction: NP_3f8,
        mask: NP_3f8
    ) -> None:
        assert (factor >= 0.0).all(), "Scale vector must be positive"
        factor = np.maximum(factor, 1e-8)
        super().__init__(
            model=model,
            about=about,
            direction=direction
        )
        self._factor: NP_3f8 = factor
        self._mask: NP_3f8 = mask

    def _get_matrix(
        self: Self,
        alpha: float
    ) -> NP_44f8:
        return ModelMatrix._matrix_from_scale(self._factor ** (self._mask * alpha))


class ModelRotateAnimation(ModelAnimation):
    __slots__ = (
        "_rotvec",
        "_mask"
    )

    def __init__(
        self: Self,
        model: Model,
        rotvec: NP_3f8,
        about: Model,
        direction: NP_3f8,
        mask: NP_3f8
    ) -> None:
        super().__init__(
            model=model,
            about=about,
            direction=direction
        )
        self._rotvec: NP_3f8 = rotvec
        self._mask: NP_3f8 = mask

    def _get_matrix(
        self: Self,
        alpha: float
    ) -> NP_44f8:
        return ModelMatrix._matrix_from_rotate(self._rotvec * (self._mask * alpha))


class ModelApplyAnimation(ModelAnimation):
    __slots__ = ("_matrix",)

    def __init__(
        self: Self,
        model: Model,
        matrix: NP_44f8,
        about: Model,
        direction: NP_3f8
    ) -> None:
        super().__init__(
            model=model,
            about=about,
            direction=direction
        )
        self._matrix: NP_44f8 = matrix

    def _get_matrix(
        self: Self,
        alpha: float
    ) -> NP_44f8:
        return (1.0 - alpha) * np.identity(4) + alpha * self._matrix



from __future__ import annotations


from typing import (
    Self,
    Unpack
)

import numpy as np

from ...animatables.animatable.animatable import AnimatableActions
from ...animatables.mesh import Mesh
from ...animatables.model import SetKwargs
from ...animatables.shape import (
    Shape,
    Triangulation
)
from ...lazy.lazy import Lazy
from ..graph_mobjects.graph_mobject import GraphMobject
from ..mesh_mobjects.mesh_mobject import MeshMobject


class ShapeMobject(MeshMobject):
    __slots__ = ()

    def __init__(
        self: Self,
        shape: Shape | None = None
    ) -> None:
        super().__init__()
        if shape is not None:
            self._shape_ = shape

    @AnimatableActions.interpolate.register_descriptor()
    @AnimatableActions.piecewise.register_descriptor()
    @Lazy.volatile()
    @staticmethod
    def _shape_() -> Shape:
        return Shape()

    @Lazy.property()
    @staticmethod
    def _mesh_(
        shape__triangulation: Triangulation
    ) -> Mesh:
        coordinates = shape__triangulation.coordinates
        faces = shape__triangulation.faces
        positions = np.concatenate((coordinates, np.zeros((len(coordinates), 1))), axis=1)
        normals = np.concatenate((np.zeros_like(coordinates), np.ones((len(coordinates), 1))), axis=1)
        return Mesh(
            positions=positions,
            normals=normals,
            uvs=coordinates,
            faces=faces
        )

    def build_stroke(
        self: Self,
        **kwargs: Unpack[SetKwargs]
    ) -> GraphMobject:
        stroke = GraphMobject()
        stroke._model_matrix_ = self._model_matrix_.copy()
        stroke._graph_ = self._shape_._graph_.copy()
        stroke.set(**kwargs)
        return stroke

    def add_strokes(
        self: Self,
        **kwargs: Unpack[SetKwargs]
    ) -> Self:
        for mobject in self.iter_descendants():
            if isinstance(mobject, ShapeMobject):
                mobject.add(mobject.build_stroke(**kwargs))
        return self



from __future__ import annotations


from typing import Self

import numpy as np

from ...animatables.graph import Graph
from ...constants.custom_typing import NP_x3f8
from .graph_mobject import GraphMobject


class Polyline(GraphMobject):
    __slots__ = ()

    def __init__(
        self: Self,
        positions: NP_x3f8
    ) -> None:
        arange = np.arange(len(positions))
        super().__init__(Graph(
            positions=positions,
            edges=np.vstack((arange[:-1], arange[1:])).T
        ))



from __future__ import annotations


import collections
import itertools
from typing import (
    Iterator,
    Self
)

import rich.box
import rich.live
import rich.table

from .toplevel import Toplevel
from .toplevel_resource import ToplevelResource


class Logger(ToplevelResource):
    __slots__ = ("_log_messages",)

    def __init__(
        self: Self
    ) -> None:
        super().__init__()
        self._log_messages: collections.deque[str] = collections.deque(maxlen=10)

    def __contextmanager__(
        self: Self
    ) -> Iterator[None]:
        Toplevel._logger = self
        with rich.live.Live(vertical_overflow="crop", get_renderable=self._get_table):
            yield
        Toplevel._logger = None

    def _get_table(
        self: Self
    ) -> rich.table.Table:
        table = rich.table.Table(
            rich.table.Column(header="Log Messages", no_wrap=True, overflow="crop", width=80),
            rich.table.Column(header="Status", no_wrap=True, overflow="crop", width=40),
            caption=f"manim3 v{__import__("manim3").__version__}",
            caption_justify="left",
            box=rich.box.ASCII
        )
        table.add_row(self._get_log_table(), self._get_status_table())
        return table

    def _get_log_table(
        self: Self
    ) -> rich.table.Table:
        log_table = rich.table.Table.grid(
            rich.table.Column(no_wrap=True, overflow="crop")
        )
        assert (maxlen := self._log_messages.maxlen) is not None
        for log_message in itertools.chain(
            self._log_messages,
            itertools.repeat("", maxlen - len(self._log_messages))
        ):
            log_table.add_row(log_message)
        return log_table

    def _get_status_table(
        self: Self
    ) -> rich.table.Table:

        def format_duration(
            duration: float
        ) -> str:
            minutes, seconds = divmod(int(duration), 60)
            return f"{minutes}:{seconds:02}"

        timer = Toplevel._get_timer()
        renderer = Toplevel._renderer
        scene = Toplevel._scene
        status_dict = {
            "Run Time": format_duration(timer._current_timestamp - timer._start_timestamp),
            "FPS": f"{timer._recorded_fps}",
            "Livestream": "-" if renderer is None else "[green]On" if renderer._livestreamer.is_livestreaming else "[red]Off",
            "Recording": "-" if renderer is None else "[green]On" if renderer._video_recorder.is_recording else "[red]Off",
            "Scene Name": "-" if scene is None else type(scene).__name__,
            "Scene Time": "-" if scene is None else format_duration(scene._scene_time)
        }

        status_table = rich.table.Table.grid(
            rich.table.Column(no_wrap=True, overflow="crop"),
            rich.table.Column(no_wrap=True, overflow="crop"),
            padding=(0, 2)
        )
        for status_key, status_value in status_dict.items():
            status_table.add_row(status_key, status_value)
        return status_table

    def log(
        self: Self,
        message: str
    ) -> None:
        self._log_messages.append(message)



from __future__ import annotations


from abc import (
    ABC,
    abstractmethod
)
from typing import Self

import attrs
import numpy as np

from ...constants.custom_typing import (
    NP_xf8,
    NP_xi4
)


@attrs.frozen(kw_only=True)
class PiecewiseInfo:
    split_alphas: NP_xf8
    concatenate_indices: NP_xi4


class Piecewiser(ABC):
    __slots__ = (
        "_backwards",
        "_n_segments"
    )

    def __init__(
        self: Self,
        n_segments: int,
        backwards: bool
    ) -> None:
        super().__init__()
        self._n_segments: int = n_segments
        self._backwards: bool = backwards

    def piecewise(
        self: Self,
        alpha: float
    ) -> PiecewiseInfo:
        n_segments = self._n_segments
        backwards = self._backwards
        start, stop = self.get_segment(alpha)
        if backwards:
            start, stop = -stop, -start
        start, stop = start % 1.0, start % 1.0 + float(np.clip(stop - start, 0.0, 1.0))

        if stop > 1.0:
            offset_0, offset_1 = stop - 1.0, start
            first_index = 0
        else:
            offset_0, offset_1 = start, stop
            first_index = 1

        split_alphas = np.column_stack(tuple(
            np.arange(n_segments, dtype=np.float64) + offset
            for offset in (offset_0, offset_1)
        )).flatten() / n_segments
        concatenate_indices = np.arange(first_index, 2 * n_segments + 1, 2)
        return PiecewiseInfo(
            split_alphas=split_alphas,
            concatenate_indices=concatenate_indices
        )

    @abstractmethod
    def get_segment(
        self: Self,
        alpha: float
    ) -> tuple[float, float]:
        pass



from __future__ import annotations
import itertools


import math
from functools import reduce
from typing import (
    Never,
    Self
)

from scipy.interpolate import BSpline

from .constants import TAU
from .custom_typing import RateType


class Rates:
    __slots__ = ()

    def __new__(
        cls: type[Self]
    ) -> Never:
        raise TypeError

    @classmethod
    def compose(
        cls: type[Self],
        *rates: RateType
    ) -> RateType:
        return lambda time: reduce(lambda t, rate: rate(t), reversed(rates), time)

    @classmethod
    def sum(
        cls: type[Self],
        *rates: RateType
    ) -> RateType:
        return lambda time: reduce(float.__add__, (rate(time) for rate in rates), 0.0)

    @classmethod
    def product(
        cls: type[Self],
        *rates: RateType
    ) -> RateType:
        return lambda time: reduce(float.__mul__, (rate(time) for rate in rates), 1.0)

    @classmethod
    def clip(
        cls: type[Self],
        rate: RateType,
        delta_time: float = 1.0,
        delta_alpha: float = 1.0,
        start_time: float = 0.0,
        start_alpha: float = 0.0
    ) -> RateType:
        return lambda time: (rate(time * delta_time + start_time) - start_alpha) / delta_alpha

    @classmethod
    def linear(
        cls: type[Self]
    ) -> RateType:
        return lambda time: time

    @classmethod
    def bezier(
        cls: type[Self],
        *anchors: float
    ) -> RateType:
        degree = len(anchors) - 1
        curve = BSpline(
            t=(*itertools.repeat(0.0, degree + 1), *itertools.repeat(1.0, degree + 1)),
            c=anchors,
            k=degree
        )
        return lambda time: float(curve(time))

    @classmethod
    def rewind(
        cls: type[Self]
    ) -> RateType:
        return lambda time: 1.0 - time

    @classmethod
    def there_and_back(
        cls: type[Self]
    ) -> RateType:
        return lambda time: 1.0 - abs(2.0 * time - 1.0)

    @classmethod
    def sinusoidal(
        cls: type[Self]
    ) -> RateType:
        return lambda time: math.sin(TAU * time)

    @classmethod
    def repeat(
        cls: type[Self],
        periods: int
    ) -> RateType:
        return lambda time: periods * time % 1.0

    # Variant rates, borrowed from `3b1b/manim`.

    @classmethod
    def smoothstep(
        cls: type[Self],
        degree: int = 1
    ) -> RateType:
        return cls.bezier(*itertools.repeat(0.0, degree + 1), *itertools.repeat(1.0, degree + 1))

    @classmethod
    def smooth(
        cls: type[Self]
    ) -> RateType:
        return cls.smoothstep(2)

    @classmethod
    def rush_from(
        cls: type[Self]
    ) -> RateType:
        return cls.clip(cls.smooth(), delta_time=0.5, delta_alpha=0.5, start_time=0.5, start_alpha=0.5)

    @classmethod
    def rush_into(
        cls: type[Self]
    ) -> RateType:
        return cls.clip(cls.smooth(), delta_time=0.5, delta_alpha=0.5)

    @classmethod
    def overshoot(
        cls: type[Self]
    ) -> RateType:
        return cls.bezier(0.0, 0.0, 1.5, 1.5, 1.0, 1.0)

    @classmethod
    def undershoot(
        cls: type[Self]
    ) -> RateType:
        return cls.bezier(0.0, 0.0, -0.5, -0.5, 1.0, 1.0)

    @classmethod
    def there_and_back_smoothly(
        cls: type[Self]
    ) -> RateType:
        return cls.compose(cls.smooth(), cls.there_and_back())

    @classmethod
    def wiggle(
        cls: type[Self]
    ) -> RateType:
        return cls.product(cls.there_and_back_smoothly(), cls.sinusoidal())
